/*
 * common.c by xenion
 *
 * Copyright (c) 2007 Dallachiesa Michele <micheleDOTdallachiesaATposteDOTit>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */


#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>
#include <grp.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <syslog.h>
#include <dirent.h>
#include "common.h"


/* globals */

static int verbose_enabled = 0;
static FILE *f = NULL;
static int syslog_enabled = 0;
static int stdout_enabled = 0;


/* extern */

extern void cleanup();


/* protos */

void sigdie(int signo);


/*******************************************/


void enable_verbose()
{
  verbose_enabled = 1;
}


void disable_verbose()
{
  verbose_enabled = 0;
}

void enable_syslog()
{
  syslog_enabled = 1;
}


void disable_syslog()
{
  syslog_enabled = 0;
}


void enable_stdout()
{
  stdout_enabled = 1;
}


void disable_stdout()
{
  stdout_enabled = 0;
}


void open_logfile(char *pathname)
{
  if (f)
    SAFE_FCLOSE(f);

  if ((f =  fopen(pathname, "a")) == NULL)
    FATAL("fopen(): %s", strerror(errno));
}


void close_logfile()
{
  if (f != NULL && f != stdout && f != stderr)
    SAFE_FCLOSE(f);
}


void
logthis(char *file, const char *function, int line, int ifverbose,int h, int n, const char *fmt, ...)
{
  static int after_h = 0;
  static char s[512] = {0, };

  va_list         ap;

  // The program must first execute the macro va_start within the body
  // of the function to initialize an object with context information.

  va_start(ap, fmt);

  if (ifverbose && verbose_enabled == 0)
    return;

  if (f == NULL)
    f = stdout;
  if (f == stdout)
    disable_stdout();


  if (after_h)
    h = 0;

  if (h)
    {
      if (verbose_enabled)
        snprintf(s+strlen(s),sizeof(s)-strlen(s),"%s %s:%d:",strtime(time(NULL)),function, line);
    }

  vsnprintf(s+strlen(s),sizeof(s)-strlen(s),fmt, ap);

  if (n)
    {
      if (strlen(s)<sizeof(s))
        strcat(s,"\n");
      else
        FATAL("line buffer full");
      fprintf(f, s);
      if (syslog_enabled)
        syslog(LOG_DAEMON|LOG_INFO, s);
      if (stdout_enabled)
        printf("%s", s);
      s[0] = '\0';
    }

  va_end(ap);

  if (n)
    after_h = 0;
  else
    after_h = 1;
}


void
fatal(char *file, const char *function, int line, const char *fmt, ...)
{
  static char s[512] = {0, }; // static per evitare allocazione di memoria, che in una funzione fatal non e' mai una buona idea
  va_list         ap;


  // The program must first execute the macro va_start within the body
  // of the function to initialize an object with context information.

  va_start(ap, fmt);

  if (f == NULL)
    f = stdout;
  if (f == stdout)
    disable_stdout();


  snprintf(s+strlen(s),sizeof(s)-strlen(s),"Fatal error at %s:%d:%s(...): ",file, line, function);
  vsnprintf(s+strlen(s),sizeof(s)-strlen(s),fmt, ap);
  snprintf(s+strlen(s),sizeof(s)-strlen(s),"; exit forced.");

  fprintf(f,"--\n%s\n--\n", s);

  if (syslog_enabled)
    syslog(LOG_DAEMON | LOG_ERR,"%s", s);

  if (stdout_enabled)
    printf("--\n%s\n--\n", s);

  va_end(ap);

  sig_unlock(); // if lock, unlock. else, nothing changes.
  sigdie(-1);
}


void
sigdie(int signo)
{
  static int loop = 0;


// if loop==1 happens, there's an infinite loop (and a bug somewhere ...). prevent it exiting. better than nothing...
  if (loop == 0)
    loop = 1;
  else
    exit(1);

// if signo == -1, it's a direct call from function fatal: less output
// messages looks better.

  if (signo != -1)
    {
      LOG(1,1,"--");
      LOG(1,1,"Caught %s signal (%d), cleaning up...", SIG_NAME(signo), signo);
      LOG(1,1,"--");
    }

  cleanup();

  LOG(1,1,"");

  // questa deve essere l'ultima cosa prima della exit.
  disable_syslog();
  close_logfile();
  disable_stdout();

  exit(signo == SIGTERM ? 0 : 1); // 0 == ok, 1 == err
}


void init_sighandlers()
{
  signal(SIGSEGV, sigdie);
  signal(SIGTERM, sigdie);
  signal(SIGINT, sigdie);
  signal(SIGUSR1, sigdie);
}


void sig_lock()
{
  signal(SIGTERM, SIG_IGN);
  signal(SIGINT, SIG_IGN);
  signal(SIGUSR1, SIG_IGN);
}


void sig_unlock()
{
  signal(SIGTERM, sigdie);
  signal(SIGINT, sigdie);
  signal(SIGUSR1, sigdie);
}


void
print_mem(u_int8_t *p, u_int32_t len, u_int32_t cols)
{
  u_int32_t off,line,col;

  printf("+ Dumping memory from %p for %d bytes %s\n",p,len, cols == 1 ? "(hex bin dec chr)" : "(hex)");

  for (off=0,col=0,line=0;off<len;off++)
    {
      if (col == 0)
        {
          printf("%s %p+%-8d",line>0?"\n|":"|",p,off);
        }

      if (cols == 1)
        {
          printf("0x%.2x %d%d%d%d%d%d%d%d %.3d %s",p[off],p[off] & 0x80 ? 1 : 0,p[off] & 0x40 ? 1 : 0,p[off] & 0x20 ? 1 : 0,p[off] & 0x10 ? 1 : 0,p[off] & 0x08 ? 1 : 0,p[off] & 0x04 ? 1 : 0,p[off] & 0x02 ? 1 : 0,p[off] & 0x01 ? 1 : 0, p[off], str_char(p[off]));
        }
      else
        printf(" %.2x",p[off]);


      if (++col >= cols)
        {
          col = 0;
          line++;
        }
    }
  printf("\n+ End\n");
}


char *str_char(unsigned char c)
{
  static char s[8];

  if (c >= 32 && 126)
    {
      sprintf(s,"'%c' ", c);
      return s;
    }

  switch (c)
    {
    case '\0':
      return "'\\0'";
    case '\r':
      return "'\\r'";
    case '\n' :
      return "'\\n'";
    default:
      sprintf(s,"?%d", c);
      return s;
    }

}


char *
strtime(time_t t)
{
  struct tm      *mytm;
  static char s[20];


  mytm = localtime(&t);
  strftime(s, 20, "%d/%m/%Y#%H:%M:%S", mytm);
  return s;
}


void drop_privs(char *user, char *group)
{
  struct passwd *p;
  struct group *g;
  int uid, gid;


  if (!user && !group)
    FATAL("(user == NULL && group == NULL)");

  if ((p = getpwnam(user)) == NULL)
    FATAL("(getpwnam(...) == NULL): user not found");

  uid = p->pw_uid;
  gid = p->pw_gid;

  if (group)
    {
      if ( (g = getgrnam(group)) == NULL)
        FATAL("(getgrnam(...) == NULL): group not found");
      gid = g->gr_gid;
    }


  if (setgid(gid) == -1)
    FATAL("setgid(...): %s",strerror(errno));

  if (setuid(uid) == -1)
    FATAL("setuid(...): %s",strerror(errno));

  if (setegid(gid) == -1)
    FATAL("setegid(...): %s",strerror(errno));

  if (seteuid(uid) == -1)
    FATAL("seteuid(...): %s",strerror(errno));
}


void daemonize()
{
  setsid();
  if (fork())
    exit(0);
}


int exists(char *pathname)
{
  struct stat st;
  int z;


  z = stat(pathname, &st);

  if (z == 0)
    return 1;

  if (errno == ENOENT)
    return 0;

  return -1; // maybe exists, maybe perm problems....
}


char *get_next_name(char *directory, char *prefix, char *suffix, int *i)
{
  int count;
  static char pathname[PATH_MAX];
  struct stat st;


  if (directory[0] != '\0' && // if directory specified...
      !isdirectory(directory))
    FATAL("'%s' is not a valid directory, check pathname and permissions", directory);

  for (count = 0;;)
    {
      snprintf(pathname, PATH_MAX, "%s/%s%d%s", directory, prefix, count, suffix);
      if (stat(pathname, &st) == -1)
        {
          if (errno == ENOENT)
            break;
          else
            {
              if (i)
                *i = -1;
              return NULL;
            }
        }
      count++;
    }

  if (i)
    *i = count;

  return pathname;
}


int isdirectory(char *pathname)
{
  struct stat st;

  if (stat(pathname, &st) != 0 || !S_ISDIR(st.st_mode))
    return 0;
  else return 1;
}


/* EOF*/

